import logging
import pathlib

import pkg_resources

from mopidy import config, ext

__version__ = pkg_resources.get_distribution("mopidy_pidi").version

logger = logging.getLogger(__name__)


class Extension(ext.Extension):

    dist_name = "Mopidy-PiDi"
    ext_name = "pidi"
    version = __version__

    @classmethod
    def get_display_types(self):
        display_types = {}
        for entry_point in pkg_resources.iter_entry_points("pidi.plugin.display"):
            try:
                plugin = entry_point.load()
                display_types[plugin.option_name] = plugin
            except (ImportError) as err:
                logger.log(
                    logging.WARN, f"Error loading display plugin {entry_point}: {err}"
                )

        return display_types

    def get_default_config(self):
        return config.read(pathlib.Path(__file__).parent / "ext.conf")

    def get_config_schema(self):
        schema = super().get_config_schema()
        schema["display"] = config.String(choices=self.get_display_types().keys())
        schema["rotation"] = config.Integer(choices=[0, 90, 180, 270])
        return schema

    def setup(self, registry):
        from .frontend import PiDiFrontend

        registry.add("frontend", PiDiFrontend)



"""
Musicbrainz related functions.
"""
import base64
import logging
import os
import time
from threading import Thread

import musicbrainzngs as mus

from .__init__ import __version__

logger = logging.getLogger(__name__)


class Brainz:
    def __init__(self, cache_dir):
        """Initialize musicbrainz."""
        mus.set_useragent(
            "python-pidi: A cover art daemon.",
            __version__,
            "https://github.com/pimoroni/mopidy-pidi",
        )

        self._cache_dir = cache_dir
        self._default_filename = os.path.join(self._cache_dir, "__default.jpg")

        self.save_album_art(self.get_default_album_art(), self._default_filename)

    def get_album_art(self, artist, album, callback=None):
        if artist is None or album is None or artist == "" or album == "":
            if callback is not None:
                return callback(self._default_filename)
            return self._default_filename

        file_name = self.get_cache_file_name(f"{artist}_{album}")

        if os.path.isfile(file_name):
            # If a cached file already exists, use it!
            if callback is not None:
                return callback(file_name)
            return file_name

        if callback is not None:

            def async_request_album_art(self, artist, album, file_name, callback):
                album_art = self.request_album_art(artist, album)

                if album_art is None:
                    # If the MusicBrainz request fails, cache the default
                    # art using this filename.
                    self.save_album_art(self.get_default_album_art(), file_name)
                    return callback(file_name)

                self.save_album_art(album_art, file_name)

                return callback(file_name)

            t_album_art = Thread(
                target=async_request_album_art,
                args=(self, artist, album, file_name, callback),
            )
            t_album_art.start()
            return t_album_art

        else:
            album_art = self.request_album_art(artist, album)

            if album_art is None:
                # If the MusicBrainz request fails, cache the default
                # art using this filename.
                self.save_album_art(self.get_default_album_art(), file_name)
                return file_name

            self.save_album_art(album_art, file_name)

            return file_name

    def save_album_art(self, data, output_file):
        with open(output_file, "wb") as f:
            f.write(data)

    def request_album_art(self, artist, album, size=500, retry_delay=5, retries=5):
        """Download the cover art."""
        try:
            data = mus.search_releases(artist=artist, release=album, limit=1)
            release_id = data["release-list"][0]["release-group"]["id"]
            logger.info("mopidy-pidi: musicbrainz using release-id: {release_id}")

            return mus.get_release_group_image_front(release_id, size=size)

        except mus.NetworkError:
            if retries == 0:
                # raise mus.NetworkError("Failure connecting to MusicBrainz.org")
                return None
            logger.info(
                f"mopidy-pidi: musicbrainz retrying download. {retries} retries left!"
            )
            time.sleep(retry_delay)
            self.request_album_art(artist, album, size=size, retries=retries - 1)

        except mus.ResponseError:
            logger.info(
                f"mopidy-pidi: musicbrainz couldn't find album art for {artist} - {album}"
            )
            return None

    def get_cache_file_name(self, file_name):
        file_name = file_name.encode("utf-8")
        file_name = base64.b64encode(file_name)
        if type(file_name) is bytes:
            file_name = file_name.decode("utf-8")
        # Ruh roh, / is a vaild Base64 character
        # but also a valid UNIX path separator!
        file_name = file_name.replace("/", "-")
        file_name = f"{file_name}.jpg"

        return os.path.join(self._cache_dir, file_name)

    def get_default_album_art(self):
        """Return binary version of default album art."""
        return base64.b64decode(
            """
iVBORw0KGgoAAAANSUhEUgAAAB4AAAAeCAMAAAAM7l6QAAAAGXRFWHRTb2Z0d2FyZQBBZG9iZSBJbWFn
ZVJlYWR5ccllPAAAAMBQTFRFBHwvBSl8d04DCQ99egJLfAMzejQGcGoAAGZ6AHN3N3wBSHwBKXwDAHlp
NQF9AHtXAFV7VwB7HgN9B30aG30FXncAAXtERwB8fQMbZQB5AUF8fRsHQ04rfQgLFlZTVzgteABiZ14F
agNiAmpoF3kaLVU4V1QVYhdFLkZIQy1MFWc/biYkKSVpLWUmLjVYcQBzJHMbeRQiBWxZBlxnOmkXDn0M
WAdnGhd5FkBlSRZfCk1rO3MMTmwJCm5FQgtwMhJydzVfDgAAAYtJREFUeNpUzeligjAQBOCNgFcVFVRQ
FC3gUU/Uingg7/9W3U1CpJOf38wGGpQ2ptPpDIcAYNv29Xrt9/utVqsJXBsfLmmzKbiYy3WZ6/XC1fyj
X8iiIOZQsFDBvFBct+1I6BcGuvUuedgIwzOfR9dI6QC6FF4I2+dsmEEURVIHA+RxVzZwfs4gi+JW3Hwi
ch5juF8ul/CcbTZxHD+ffFqwrGDB32z2+9/n6/VCqw1qwMZMFh6Ph+/7C2RUJAowGWqlqb9eLCa/y2/M
f2YsZWl6WK8nk+VSOTBN05iGemO73e5w+JnNZpVlRQYIKTcM+g/xtiq1BloR5Dy/3++r7ba6rWLkmmLd
LCvP8zfqCp0zNYgtepZlmu93kiCfTifP87iDNK5OkiSBbpyEe1WPs0DTdJxeEAQr3TCUgyXUQnR6ySgI
dJy7rjclV8y3PdS5jm647nRKDVBIOjoSG4KpAOpfB3V0nM/LjmyapXHBriscylrwx0FpiQ11Hf6PyXX5
ORWAoxqr44Y4/ifAAPd/TAMIg8r1AAAAAElFTkSuQmCC"""
        )



from unittest import mock

from mopidy_pidi import Extension
from mopidy_pidi import frontend as frontend_lib


def test_get_default_config():
    ext = Extension()

    config = ext.get_default_config()

    assert "[pidi]" in config
    assert "enabled = true" in config


def test_get_config_schema():
    ext = Extension()

    schema = ext.get_config_schema()

    assert "display" in schema


def test_setup():
    ext = Extension()
    registry = mock.Mock()

    ext.setup(registry)

    registry.add.assert_called_once_with("frontend", frontend_lib.PiDiFrontend)



import pkg_resources
import pykka
from mopidy import core

import pytest
from mopidy_pidi import frontend as frontend_lib

from . import dummy_audio, dummy_backend, dummy_mixer


def stop_mopidy_core():
    pykka.ActorRegistry.stop_all()


@pytest.fixture(scope="session", autouse=True)
def cleanup(request):
    request.addfinalizer(stop_mopidy_core)


@pytest.fixture
def frontend():
    mixer = dummy_mixer.create_proxy()
    audio = dummy_audio.create_proxy()
    backend = dummy_backend.create_proxy(audio=audio)
    dummy_core = core.Core.start(audio=audio, mixer=mixer, backends=[backend]).proxy()

    distribution = pkg_resources.Distribution(__file__)
    endpoint = pkg_resources.EntryPoint.parse(
        "dummy = mopidy_pidi.plugin:DisplayDummy", dist=distribution
    )
    distribution._ep_map = {"pidi.plugin.display": {"dummy": endpoint}}
    pkg_resources.working_set.add(distribution, "dummy")

    config = {"pidi": {"display": "dummy"}, "core": {"data_dir": "/tmp"}}

    return frontend_lib.PiDiFrontend(config, dummy_core)


def test_on_start(frontend):
    frontend.on_start()
    frontend.on_stop()


def test_options_changed(frontend):
    frontend.on_start()
    frontend.options_changed()
    frontend.on_stop()



import logging
import os
import threading
import time

import pykka
import requests
from mopidy import core

import netifaces

from . import Extension
from .brainz import Brainz

logger = logging.getLogger(__name__)


class PiDiConfig:
    def __init__(self, config=None):
        self.rotation = config.get("rotation", 90)
        self.spi_port = 0
        self.spi_chip_select_pin = 1
        self.spi_data_command_pin = 9
        self.spi_speed_mhz = 80
        self.backlight_pin = 13
        self.size = 240
        self.blur_album_art = True


class PiDiFrontend(pykka.ThreadingActor, core.CoreListener):
    def __init__(self, config, core):
        super().__init__()
        self.core = core
        self.config = config
        self.current_track = None

    def on_start(self):
        self.display = PiDi(self.config)
        self.display.start()
        self.display.update(volume=self.core.mixer.get_volume().get())
        if "http" in self.config:
            ifaces = netifaces.interfaces()
            ifaces.remove("lo")

            http = self.config["http"]
            if http.get("enabled", False):
                hostname = http.get("hostname", "127.0.0.1")
                port = http.get("port", 6680)
                if hostname in ["::", "0.0.0.0"]:
                    family = (
                        netifaces.AF_INET6 if hostname == "::" else netifaces.AF_INET
                    )
                    for iface in ifaces:
                        hostname = self.get_ifaddress(iface, family)
                        if hostname is not None:
                            break
                if hostname is not None:
                    self.display.update(
                        title=f"Visit http://{hostname}:{port} to select content."
                    )
                    self.display.update_album_art(art="")

    def on_stop(self):
        self.display.stop()
        self.display = None

    def get_ifaddress(self, iface, family):
        try:
            return netifaces.ifaddresses(iface)[family][0]["addr"]
        except (IndexError, KeyError):
            return None

    def mute_changed(self, mute):
        pass

    def options_changed(self):
        self.display.update(
            shuffle=self.core.tracklist.get_random(),
            repeat=self.core.tracklist.get_repeat(),
        )

    def playlist_changed(self, playlist):
        pass

    def playlist_deleted(self, playlist):
        pass

    def playlists_loaded(self):
        pass

    def seeked(self, time_position):
        self.update_elapsed(time_position)

    def stream_title_changed(self, title):
        self.display.update(title=title)

    def track_playback_ended(self, tl_track, time_position):
        self.update_elapsed(time_position)
        self.display.update(state="pause")

    def track_playback_paused(self, tl_track, time_position):
        self.update_elapsed(time_position)
        self.display.update(state="pause")

    def track_playback_resumed(self, tl_track, time_position):
        self.update_elapsed(time_position)
        self.display.update(state="play")

    def track_playback_started(self, tl_track):
        self.update_track(tl_track.track, 0)
        self.display.update(state="play")

    def update_elapsed(self, time_position):
        self.display.update(elapsed=float(time_position))

    def update_track(self, track, time_position=None):
        if track is None:
            track = self.core.playback.get_current_track().get()

        title = ""
        album = ""
        artist = ""

        if track.name is not None:
            title = track.name

        if track.album is not None and track.album.name is not None:
            album = track.album.name

        if track.artists is not None:
            artist = ", ".join([artist.name for artist in track.artists])

        self.display.update(title=title, album=album, artist=artist)

        if time_position is not None:
            length = track.length
            # Default to 60s long and loop the transport bar
            if length is None:
                length = 60
                time_position %= length

            self.display.update(elapsed=float(time_position), length=float(length))

        art = None
        track_images = self.core.library.get_images([track.uri]).get()
        if track.uri in track_images:
            track_images = track_images[track.uri]
            if len(track_images) == 1:
                art = track_images[0].uri
            else:
                for image in track_images:
                    if image.width is None or image.height is None:
                        continue
                    if image.height >= 240 and image.width >= 240:
                        art = image.uri

        self.display.update_album_art(art=art)

    def tracklist_changed(self):
        pass

    def volume_changed(self, volume):
        if volume is None:
            return

        self.display.update(volume=volume)


class PiDi:
    def __init__(self, config):
        self.config = config
        self.cache_dir = Extension.get_data_dir(config)
        self.display_config = PiDiConfig(config["pidi"])
        self.display_class = Extension.get_display_types()[
            self.config["pidi"]["display"]
        ]

        self._brainz = Brainz(cache_dir=self.cache_dir)
        self._display = self.display_class(self.display_config)
        self._running = threading.Event()
        self._delay = 1.0 / 30
        self._thread = None

        self.shuffle = False
        self.repeat = False
        self.state = "stop"
        self.volume = 100
        self.progress = 0
        self.elapsed = 0
        self.length = 0
        self.title = ""
        self.album = ""
        self.artist = ""
        self._last_progress_update = time.time()
        self._last_progress_value = 0
        self._last_art = ""

    def start(self):
        if self._thread is not None:
            return

        self._running = threading.Event()
        self._running.set()
        self._thread = threading.Thread(target=self._loop)
        self._thread.start()

    def stop(self):
        self._running.clear()
        self._thread.join()
        self._thread = None
        self._display.stop()

    def _handle_album_art(self, art):
        if art != self._last_art:
            self._display.update_album_art(art)
            self._last_art = art

    def update_album_art(self, art=None):
        _album = self.title if self.album is None or self.album == "" else self.album

        if art is not None:
            if os.path.isfile(art):
                # Art is already a locally cached file we can use
                self._handle_album_art(art)
                return

            elif art.startswith("http://") or art.startswith("https://"):
                file_name = self._brainz.get_cache_file_name(art)

                if os.path.isfile(file_name):
                    # If a cached file already exists, use it!
                    self._handle_album_art(file_name)
                    return

                else:
                    # Otherwise, request the URL and save it!
                    response = requests.get(art)
                    if response.status_code == 200:
                        self._brainz.save_album_art(response.content, file_name)
                        self._handle_album_art(file_name)
                        return

        art = self._brainz.get_album_art(self.artist, _album, self._handle_album_art)

    def update(self, **kwargs):
        self.shuffle = kwargs.get("shuffle", self.shuffle)
        self.repeat = kwargs.get("repeat", self.repeat)
        self.state = kwargs.get("state", self.state)
        self.volume = kwargs.get("volume", self.volume)
        # self.progress = kwargs.get('progress', self.progress)
        self.elapsed = kwargs.get("elapsed", self.elapsed)
        self.length = kwargs.get("length", self.length)
        self.title = kwargs.get("title", self.title)
        self.album = kwargs.get("album", self.album)
        self.artist = kwargs.get("artist", self.artist)

        if "elapsed" in kwargs:
            if "length" in kwargs:
                self.progress = float(self.elapsed) / float(self.length)
            self._last_elapsed_update = time.time()
            self._last_elapsed_value = kwargs["elapsed"]

    def _loop(self):
        while self._running.is_set():
            if self.state == "play":
                t_elapsed_ms = (time.time() - self._last_elapsed_update) * 1000
                self.elapsed = float(self._last_elapsed_value + t_elapsed_ms)
                self.progress = self.elapsed / self.length
            self._display.update_overlay(
                self.shuffle,
                self.repeat,
                self.state,
                self.volume,
                self.progress,
                self.elapsed,
                self.title,
                self.album,
                self.artist,
            )

            self._display.redraw()
            time.sleep(self._delay)



class Display:
    """Base class to represent a Pirate Display display output."""

    def __init__(self, args=None):
        """Initialise a new display."""
        self._size = args.size
        self._title = ""
        self._shuffle = False
        self._repeat = False
        self._state = ""
        self._volume = 0
        self._progress = 0
        self._elapsed = 0

        self._title = ""
        self._album = ""
        self._artist = ""

    def stop(self):
        pass

    def update_album_art(self, input_file):
        """Update the display album art."""
        raise NotImplementedError

    def update_overlay(
        self, shuffle, repeat, state, volume, progress, elapsed, title, album, artist
    ):
        """Update the display transport information."""
        self._shuffle = shuffle
        self._repeat = repeat
        self._state = state
        self._volume = float(volume)
        self._progress = progress
        self._elapsed = elapsed
        self._title = title
        self._album = album
        self._artist = artist

    def redraw(self):
        """Redraw the display."""
        raise NotImplementedError

    def add_args(argparse):
        """Expand argparse instance with display-specific args."""


class DisplayDummy(Display):
    """Dummy display for use in texting."""

    option_name = "dummy"

    def update_album_art(self, input_file):
        pass

    def redraw(self):
        pass
